# Generated by Django 2.0rc1 on 2017-11-19 08:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0008_auto_20171115_1601")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Routine / Scheduled"),
                    ("unscheduled", "Unscheduled"),
                ],
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
    ]
